"""Test the pydantic schema"""

from copy import deepcopy
from typing import get_args
from uuid import UUID

import pytest
from pydantic import BaseModel, ValidationError

from fmu.datamodels import FmuResults
from fmu.datamodels.fmu_results import enums
from fmu.datamodels.fmu_results.fields import Aggregation, Realization


def test_for_optional_fields_without_default(
    pydantic_models_from_root: set[type[BaseModel]],
) -> None:
    """Test that all optional fields have a default value."""

    optionals_without_default = []
    for model in pydantic_models_from_root:
        for field_name, field_info in model.model_fields.items():
            if (
                type(None) in get_args(field_info.annotation)
                and field_info.is_required()
            ):
                optionals_without_default.append(
                    f"{model.__module__}.{model.__name__}.{field_name}"
                )

    assert not optionals_without_default


def test_fmu_field(case_metadata: dict, field_region_metadata: dict) -> None:
    """Asserting validation failure when illegal contents in fmu block."""

    # assert validation as-is
    FmuResults.model_validate(case_metadata)
    FmuResults.model_validate(field_region_metadata)

    # assert validation error when "fmu" is missing
    _example = deepcopy(case_metadata)
    del _example["fmu"]
    with pytest.raises(ValidationError):
        FmuResults.model_validate(_example)

    _example = deepcopy(field_region_metadata)
    del _example["fmu"]
    with pytest.raises(ValidationError):
        FmuResults.model_validate(_example)

    # assert validation error when "fmu.model" is missing
    _example = deepcopy(case_metadata)
    del _example["fmu"]["model"]
    with pytest.raises(ValidationError):
        FmuResults.model_validate(_example)

    _example = deepcopy(field_region_metadata)
    del _example["fmu"]["model"]
    with pytest.raises(ValidationError):
        FmuResults.model_validate(_example)


def test_masterdata_smda_field(
    case_metadata: dict, field_outline_metadata: dict
) -> None:
    """Test schema logic for masterdata.smda."""

    for metadata in [case_metadata, field_outline_metadata]:
        # assert validation as-is
        FmuResults.model_validate(metadata)

        # assert validation error when masterdata block is missing
        _example = deepcopy(metadata)
        del _example["masterdata"]
        with pytest.raises(ValidationError):
            FmuResults.model_validate(_example)

        # assert validation error when masterdata.smda is missing
        _example = deepcopy(metadata)
        del _example["masterdata"]["smda"]
        with pytest.raises(ValidationError):
            FmuResults.model_validate(_example)

        # assert validation error when missing attribute
        for block in [
            "country",
            "discovery",
            "field",
            "coordinate_system",
            "stratigraphic_column",
        ]:
            _example = deepcopy(metadata)
            del _example["masterdata"]["smda"][block]
            with pytest.raises(ValidationError):
                FmuResults.model_validate(_example)

        # assert validation error if not correct type
        for block, type_ in [
            ("country", list),
            ("discovery", list),
            ("coordinate_system", dict),
            ("stratigraphic_column", dict),
        ]:
            _example = deepcopy(metadata)
            assert isinstance(_example["masterdata"]["smda"][block], type_)

            _example["masterdata"]["smda"][block] = "somestring"

            with pytest.raises(ValidationError):
                FmuResults.model_validate(_example)


def test_file_field(fluid_contact_metadata: dict) -> None:
    """Test variations on the file block."""

    # assert validation as-is
    FmuResults.model_validate(fluid_contact_metadata)

    # shall validate without absolute_path
    del fluid_contact_metadata["file"]["absolute_path"]
    FmuResults.model_validate(fluid_contact_metadata)

    # shall not validate when md5 checksum is not a string
    valid_checksum = fluid_contact_metadata["file"]["checksum_md5"]
    fluid_contact_metadata["file"]["checksum_md5"] = 123.4
    with pytest.raises(ValidationError):
        FmuResults.model_validate(fluid_contact_metadata)

    # shall not validate without checksum_md5
    del fluid_contact_metadata["file"]["checksum_md5"]
    with pytest.raises(ValidationError):
        FmuResults.model_validate(fluid_contact_metadata)

    # shall validate when checksum is put back in
    fluid_contact_metadata["file"]["checksum_md5"] = valid_checksum
    FmuResults.model_validate(fluid_contact_metadata)

    # shall not validate without relative_path
    del fluid_contact_metadata["file"]["relative_path"]
    with pytest.raises(ValidationError):
        FmuResults.model_validate(fluid_contact_metadata)


def test_aggregation_and_realalization_field(fluid_contact_metadata: dict) -> None:
    """Test that fmu.realization and fmu.aggregation
    are not allowed at the same time."""

    # assert validation as-is
    FmuResults.model_validate(fluid_contact_metadata)

    # add realization block to test data
    fluid_contact_metadata["fmu"]["realization"] = Realization(
        id=0,
        name="realization-0",
        uuid=UUID("00000000-0000-0000-0000-000000000000"),
        is_reference=None,
    ).model_dump(mode="json", exclude_none=True, by_alias=True)

    # check that assumptions for the test is true
    assert "realization" in fluid_contact_metadata["fmu"]
    assert "aggregation" not in fluid_contact_metadata["fmu"]

    # assert validation with realization set
    FmuResults.model_validate(fluid_contact_metadata)

    # add aggregation block to test data
    fluid_contact_metadata["fmu"]["realization"] = Aggregation(
        id=UUID("15ce3b84-766f-4c93-9050-b154861f9100"),
        operation="std",
        realization_ids=[0, 1, 9],
    ).model_dump(mode="json", exclude_none=True, by_alias=True)

    # assert validation fail when both realization and aggregation are set
    with pytest.raises(ValidationError):
        FmuResults.model_validate(fluid_contact_metadata)


@pytest.mark.parametrize("classification", enums.Classification)
def test_access_field(
    case_metadata: dict, seismic_metadata: dict, classification: dict
) -> None:
    """Test the classification of individual files."""

    # assert valid classification validates
    case_metadata["access"]["classification"] = classification
    FmuResults.model_validate(case_metadata)

    seismic_metadata["access"]["ssdl"]["access_level"] = classification
    FmuResults.model_validate(seismic_metadata)

    # assert erroneous value does not validate
    case_metadata["access"]["classification"] = "open"
    with pytest.raises(ValidationError):
        FmuResults.model_validate(case_metadata)

    seismic_metadata["access"]["ssdl"]["access_level"] = "open"
    with pytest.raises(ValidationError):
        FmuResults.model_validate(seismic_metadata)



from .field_outline import FieldOutlineResult, FieldOutlineSchema
from .fluid_contact_outline import FluidContactOutlineResult, FluidContactOutlineSchema
from .inplace_volumes import InplaceVolumesResult, InplaceVolumesSchema
from .structure_depth_fault_lines import (
    StructureDepthFaultLinesResult,
    StructureDepthFaultLinesSchema,
)

__all__ = [
    "FieldOutlineResult",
    "FieldOutlineSchema",
    "InplaceVolumesResult",
    "InplaceVolumesSchema",
    "StructureDepthFaultLinesSchema",
    "StructureDepthFaultLinesResult",
    "FluidContactOutlineSchema",
    "FluidContactOutlineResult",
]



from __future__ import annotations

from enum import Enum, IntEnum, StrEnum


class Classification(StrEnum):
    """The security classification for a given data object."""

    asset = "asset"
    # Deprecated. This value should be undocumented and eventually removed.

    internal = "internal"
    """Grants access to all users with ``READ`` access to the asset.

    The ``READ`` role is an access role defined by the asset's Unix and Sumo groups.
    This is the default for most data.
    """

    restricted = "restricted"
    """Grants access to all users with ``WRITE`` access to the asset.

    The ``WRITE`` role is an access role defined by the asset's Unix and Sumo groups.
    This is the default for some sensitive data, like volumes, but in general must be
    explicitly set when restricted access is desired.
    """


class AxisOrientation(IntEnum):
    """The axis orientation for a given data object."""

    normal = 1
    """A coordinate system with Z-axis positive downwards."""

    flipped = -1
    """A coordinate system with Z-axis positive upwards."""


class Content(StrEnum):
    """The content type of a given data object."""

    depth = "depth"
    """A data object representing depth values.

    Typically provided as an ``xtgeo.RegularSurface`` or ``xtgeo.Grid`` for export.
    """

    facies_thickness = "facies_thickness"
    """Thickness map representing facies thickness, derived from a 3D grid.

    Typically provided as an ``xtgeo.RegularSurface`` for export.
    """

    fault_lines = "fault_lines"
    """Intersections between fault planes and horizons.

    Typically provided as an ``xtgeo.Polygons`` for export.
    """

    fault_surface = "fault_surface"
    """A surface representing a fault plane.

    Typically provided either as an RMS FaultRoom GeoJSON surface or an fmu-dataio
    ``TSurfData`` for export.
    """

    fault_properties = "fault_properties"
    """Properties, such as permeability and porosity, on a fault.

    Typically provided as a GeoJSON file derived from RMS FaultRoom for export.
    """

    field_outline = "field_outline"
    """Polygons representing the outline of a field, initial (static) conditions.

    Typically provided as an ``xtgeo.Polygons`` for export.
    """

    field_region = "field_region"
    """Delineated or named region within a field.

    Typically provided as an ``xtgeo.Polygons`` for export.
    """

    fluid_contact = "fluid_contact"
    """Depth surface representing a fluid contact used per realization.

    Typically provided as an ``xtgeo.RegularSurface`` for export.
    """

    khproduct = "khproduct"
    """The product of permeability (k) and reservoir thickness (h).

    Typically provided as an ``xtgeo.RegularSurface`` for export.
    """

    lift_curves = "lift_curves"
    """Table representing the relationship between production rates and pressures.

    Typically provided as a Pandas ``Dataframe`` for export.
    """

    named_area = "named_area"
    """A named area within a field that is _not_ a region.

    Typically provided as an ``xtgeo.Polygons`` for export.
    """

    parameters = "parameters"
    """The ERT parameters generated for the realization.

    Typically provided as a Pandas ``Dataframe`` for export.

    .. tip::

       You should not export this manually. This is done automatically by `SUMO_UPLOADER
       <https://fmu-sumo-uploader.readthedocs.io/en/latest/>`_ when uploading files.

    """

    pinchout = "pinchout"
    """Polygons designating a pinchout.

    Typically provided as an ``xtgeo.Polygons`` for export.
    """

    property = "property"
    """A property, like permeability or porosity, belonging to a 3D grid.

    Typically provided as an ``xtgeo.GridProperty``.

    .. tip::

       This content type requires additional input in the ``content_metadata`` field.

       Grid property data handling is still immature. More comprehensive data
       categorization will come in the future.

    """

    pvt = "pvt"
    """Tabular pressure-volume-temperature data.

    Typically provided as a Pandas ``Dataframe`` for export.

    .. tip::

       You should not export this manually. Use `SIM2SUMO
       <https://fmu-sumo-sim2sumo.readthedocs.io/en/latest/>`_.

    """

    regions = "regions"
    """Distinct areas within the field that have different characteristics.

    Examples may be volume regions or contact regions.

    Typically provided as an ``xtgeo.Polygons`` or ``xtgeo.GridProperty``.
    """

    relperm = "relperm"
    """Tabular relative permeability data.

    Typically provided as a Pandas ``Dataframe`` for export.

    .. tip::

       You should not export this manually. Use `SIM2SUMO
       <https://fmu-sumo-sim2sumo.readthedocs.io/en/latest/>`_.

    """

    rft = "rft"
    """Tabular reservoir formation tests data.

    .. tip::

       You should not export this manually. Use `SIM2SUMO
       <https://fmu-sumo-sim2sumo.readthedocs.io/en/latest/>`_.

    """

    seismic = "seismic"
    """Data that is seismic in nature, including seismic cubes and surface data derived
    from seismic cubes.

    Typically provided as an ``xtgeo.Cube``, ``xtgeo.RegularSurface``, or other.

    .. tip::

       This content type requires additional input in the ``content_metadata`` field.

       Seismic data handling is still immature. More comprehensive data categorization
       will come in the future.

    """

    simulationtimeseries = "simulationtimeseries"
    """Time-series data generated by a reservoir simulator like OPM Flow or Eclipse.

    For example, a summary file parsed into a Pandas ``Dataframe`` by
    `res2df <https://equinor.github.io/res2df>`_.

    .. tip::

       You should not export this manually. Use `SIM2SUMO
       <https://fmu-sumo-sim2sumo.readthedocs.io/en/latest/>`_.

    """

    subcrop = "subcrop"
    """Surface or polygon representing a subcrop area.

    Typically provided as an ``xtgeo.RegularSurface`` or ``xtgeo.Polygons`` for export.
    """

    thickness = "thickness"
    """A thickness map.

    Typically provided as an ``xtgeo.RegularSurface`` for export.
    """

    time = "time"
    """A seismic time surface or seismic cube in time domain.

    Typically provided as an ``xtgeo.RegularSurface`` or ``xtgeo.Cube``.
    """

    timeseries = "timeseries"
    # Deprecated. Use "simlationtimeseries" ?

    transmissibilities = "transmissibilities"
    """Tabular data containing transmissibilities (neighbour and
    non-neigbor-connections).

    Typically provided as a Pandas ``Dataframe``.

    .. tip::

       You should not export this manually. Use `SIM2SUMO
       <https://fmu-sumo-sim2sumo.readthedocs.io/en/latest/>`_.

    """

    velocity = "velocity"
    """A seismic velocity map represented as a regular surface or a cube.

    Typically provided as an ``xtgeo.RegularSurface`` or ``xtgeo.Cube`` for export.
    """

    volumes = "volumes"
    """Tabulated inplace volumes per grid, initial (static) conditions.

    Typically provided as a Pandas ``Dataframe``.
    """

    wellpicks = "wellpicks"
    """Tabular data representing wellpicks.

    Typically provided as a Pandas ``Dataframe``.
    """

    @classmethod
    def _missing_(cls: type[Content], value: object) -> None:
        raise ValueError(
            f"Invalid 'content' {value=}. Valid entries are {[m.value for m in cls]}"
        )


class ErtSimulationMode(str, Enum):
    """The simulation mode ert was run in. These definitions come from
    `ert.mode_definitions`."""

    ensemble_experiment = "ensemble_experiment"
    ensemble_information_filter = "ensemble_information_filter"
    ensemble_smoother = "ensemble_smoother"
    es_mda = "es_mda"
    evaluate_ensemble = "evaluate_ensemble"
    manual_update = "manual_update"
    test_run = "test_run"
    workflow = "workflow"


class MetadataClass(StrEnum):
    """Base class for objects by FMU convention or standards."""


class ObjectMetadataClass(MetadataClass):
    """The class of a data object (typically originating from an RMS model)."""

    surface = "surface"
    table = "table"
    cpgrid = "cpgrid"
    cpgrid_property = "cpgrid_property"
    polygons = "polygons"
    cube = "cube"
    well = "well"
    points = "points"
    dictionary = "dictionary"


class FMUResultsMetadataClass(MetadataClass):
    """The class of an FMU results object."""

    case = "case"
    realization = "realization"
    iteration = "iteration"
    ensemble = "ensemble"


class Layout(StrEnum):
    """The layout of a given data object."""

    regular = "regular"
    unset = "unset"
    cornerpoint = "cornerpoint"
    table = "table"
    dictionary = "dictionary"
    faultroom_triangulated = "faultroom_triangulated"
    triangulated = "triangulated"


class FMUContext(str, Enum):
    """The context in which FMU was being run when data were generated."""

    case = "case"
    iteration = "iteration"
    ensemble = "ensemble"
    realization = "realization"


class VerticalDomain(StrEnum):
    depth = "depth"
    """In the domain of depth."""

    time = "time"
    """In the domain of time."""


class DomainReference(StrEnum):
    msl = "msl"
    """In reference to Mean Sea Level."""

    sb = "sb"
    """In reference to Sea Bottom."""

    rkb = "rkb"
    """In reference to Rotary Kelly Bushing (RKB)."""


class TrackLogEventType(StrEnum):
    """The type of event being logged"""

    created = "created"
    """The initial data was created."""

    updated = "updated"
    """Data was updated."""

    merged = "merged"
    """Data was merged."""


class FluidContactType(StrEnum):
    """The type of fluid contact."""

    fgl = "fgl"
    """Free gas level."""

    fwl = "fwl"
    """Free water level."""

    goc = "goc"
    """Gas-oil contact."""

    gwc = "gwc"
    """Gas-water contact."""

    owc = "owc"
    """Oil-water contact."""


class FileFormat(StrEnum):
    """The format of a given data object."""

    parquet = "parquet"
    json = "json"
    csv = "csv"
    csv_xtgeo = "csv|xtgeo"
    irap_ascii = "irap_ascii"
    irap_binary = "irap_binary"
    roff = "roff"
    segy = "segy"
    openvds = "openvds"
    tsurf = "tsurf"



from __future__ import annotations

from pydantic import BaseModel, Field

from . import enums


class RowColumn(BaseModel):
    """Specifies the number of rows and columns in a regular surface object."""

    nrow: int = Field(gt=0)
    """The number of rows."""

    ncol: int = Field(gt=0)
    """The number of columns."""


class RowColumnLayer(RowColumn):
    """Specifies the number of rows, columns, and layers in grid object."""

    nlay: int = Field(gt=0)
    """The number of layers."""


class SurfaceSpecification(RowColumn):
    """Specifies relevant values describing a regular surface object."""

    rotation: float = Field(allow_inf_nan=False)
    """Rotation angle in degrees."""

    undef: float = Field(allow_inf_nan=False)
    """Value representing undefined data."""

    xinc: float = Field(gt=0, allow_inf_nan=False)
    """Increment along the x-axis."""

    yinc: float = Field(gt=0, allow_inf_nan=False)
    """Increment along the y-axis."""

    xori: float = Field(allow_inf_nan=False)
    """Origin along the x-axis."""

    yflip: enums.AxisOrientation
    """Flip along the y-axis, -1 or 1."""

    yori: float = Field(allow_inf_nan=False)
    """Origin along the y-axis."""


# TODO: Base on TableSpecification when we only support table export format for points
class PointSpecification(BaseModel):
    """Specifies relevant values describing an xyz points object."""

    attributes: list[str] | None = Field(default=None)
    """List of columns present in a table."""

    size: int = Field(ge=0, examples=[1, 9999])
    """Size of data object."""

    columns: list[str] | None = Field(default=None)
    """List of columns present in a table."""

    num_columns: int | None = Field(default=None, ge=0, examples=[1, 9999])
    """The number of columns in a table."""

    num_rows: int | None = Field(default=None, ge=0, examples=[1, 9999])
    """The number of rows in a table.."""


class TableSpecification(BaseModel):
    """Specifies relevant values describing a generic tabular data object."""

    columns: list[str]
    """List of columns present in a table."""

    num_columns: int = Field(ge=0, examples=[1, 9999])
    """The number of columns in a table."""

    num_rows: int = Field(ge=0, examples=[1, 9999])
    """The number of rows in a table.."""

    size: int = Field(ge=0, examples=[1, 9999])
    """The total size of the table, i.e. `rows x cols`."""


class CPGridSpecification(RowColumnLayer):
    """Specifies relevant values describing a corner point grid object."""

    xshift: float = Field(allow_inf_nan=False)
    """Shift along the x-axis."""

    yshift: float = Field(allow_inf_nan=False)
    """Shift along the y-axis."""

    zshift: float = Field(allow_inf_nan=False)
    """Shift along the z-axis."""

    xscale: float = Field(allow_inf_nan=False)
    """Scaling factor for the x-axis."""

    yscale: float = Field(allow_inf_nan=False)
    """Scaling factor for the y-axis."""

    zscale: float = Field(allow_inf_nan=False)
    """Scaling factor for the z-axis."""

    zonation: list[ZoneDefinition] | None = Field(default=None)
    """
    Zone names and corresponding layer index ranges. The list is ordered from
    shallowest to deepest zone. Note the layer indices are zero-based; add 1 to
    convert to corresponding layer number.
    """


class ZoneDefinition(BaseModel):
    """Zone name and corresponding layer index min/max"""

    name: str
    """Name of zone"""

    min_layer_idx: int = Field(ge=0)
    """Minimum layer index for the zone"""

    max_layer_idx: int = Field(ge=0)
    """Maximum layer index for the zone"""


class CPGridPropertySpecification(RowColumnLayer):
    """Specifies relevant values describing a corner point grid property object."""


# TODO: Base on TableSpecification when we only support table export format for polygons
class PolygonsSpecification(BaseModel):
    """Specifies relevant values describing a polygon object."""

    npolys: int = Field(ge=0)
    """The number of individual polygons in the data object."""

    columns: list[str] | None = Field(default=None)
    """List of columns present in a table."""

    num_columns: int | None = Field(default=None, ge=0, examples=[1, 9999])
    """The number of columns in a table."""

    num_rows: int | None = Field(default=None, ge=0, examples=[1, 9999])
    """The number of rows in a table.."""

    size: int | None = Field(default=None, ge=0, examples=[1, 9999])
    """The total size of the table, i.e. `rows x cols`."""


class FaultRoomSurfaceSpecification(BaseModel):
    """Specifies relevant values describing a Faultroom surface object."""

    horizons: list[str]
    """List of horizon names."""

    faults: list[str]
    """Names of faults."""

    juxtaposition_hw: list[str]
    """List of zones included in hangingwall juxtaposition."""

    juxtaposition_fw: list[str]
    """List of zones included in footwall juxtaposition."""

    properties: list[str]
    """List of properties along fault plane."""

    name: str
    """A name id of the faultroom usage."""


class TriangulatedSurfaceSpecification(BaseModel):
    """Specifies relevant values describing a triangulated surface."""

    num_vertices: int = Field(ge=0)
    """The number of vertices."""

    num_triangles: int = Field(ge=0)
    """The number of triangles."""


class CubeSpecification(SurfaceSpecification):
    """Specifies relevant values describing a cube object, i.e. a seismic cube."""

    nlay: int = Field(gt=0)
    """The number of layers."""

    xinc: float = Field(gt=0, allow_inf_nan=False)
    """Increment along the x-axis."""

    yinc: float = Field(gt=0, allow_inf_nan=False)
    """Increment along the y-axis."""

    zinc: float = Field(gt=0, allow_inf_nan=False)
    """Increment along the z-axis."""

    xori: float = Field(allow_inf_nan=False)
    """Origin along the x-axis."""

    yori: float = Field(allow_inf_nan=False)
    """Origin along the y-axis."""

    zori: float = Field(allow_inf_nan=False)
    """Origin along the z-axis."""

    yflip: enums.AxisOrientation
    """Flip along the y-axis, -1 or 1."""

    zflip: enums.AxisOrientation
    """Flip along the z-axis, -1 or 1."""

    rotation: float = Field(allow_inf_nan=False)
    """Rotation angle in degrees."""

    undef: float
    """Value representing undefined data."""


AnySpecification = (
    CPGridPropertySpecification
    | CPGridSpecification
    | FaultRoomSurfaceSpecification
    | PointSpecification
    | CubeSpecification
    | PolygonsSpecification
    | SurfaceSpecification
    | TriangulatedSurfaceSpecification
    | TableSpecification
)
